"""Provide exception classes for the stare package."""

import sys
import json

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError

if sys.version_info[0] == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse


class GlanceDBException(Exception):
    """Base exception class for exceptions that occur within this package."""


class InvalidInvocation(GlanceDBException):
    """Indicate that the code to execute cannot be completed."""


class ResponseException(GlanceDBException):
    """Indicate that there was an error with the completed HTTP request."""

    def __init__(self, response, additional_message=None):
        """Initialize a ResponseException instance.
        :param response: A requests.response instance.
        """
        self.response = response
        message = 'received {} HTTP response'.format(response.status_code)
        try:
            if additional_message is None:
                additional_message = json.dumps(response.json(), indent=2)

        except JSONDecodeError:
            additional_message = response.text.strip()

        if additional_message:
            message = '{}. The following details may help:\n{}'.format(
                message, additional_message
            )
        super(ResponseException, self).__init__(message)


class BadJSON(ResponseException):
    """Indicate the response did not contain valid JSON."""


class BadRequest(ResponseException):
    """Indicate invalid parameters for the request."""


class Conflict(ResponseException):
    """Indicate a conflicting change in the target resource."""


class Forbidden(ResponseException):
    """Indicate the authentication is not permitted for the request."""

    def __init__(self, response):
        additional_message = None
        try:
            additional_message = (
                response.json()
                .get('uuAppErrorMap', {})
                .get('uu-app-workspace/authorization/userIsNotAuthorized', {})
                .get('message', None)
            )
        except JSONDecodeError:
            pass
        super(Forbidden, self).__init__(response, additional_message)


class NotFound(ResponseException):
    """Indicate that the requested URL was not found."""


class Redirect(ResponseException):
    """Indicate the request resulted in a redirect.

    This class adds the attribute ``path``, which is the path to which the
    response redirects.

    """

    def __init__(self, response):
        """Initialize a Redirect exception instance.

        :param response: A requests.response instance containing a location
        header.

        """
        path = urlparse(response.headers['location']).path
        self.path = path[:-5] if path.endswith('.json') else path
        self.response = response
        GlanceDBException.__init__(self, 'Redirect to {}'.format(self.path))


class ServerError(ResponseException):
    """Indicate issues on the server end preventing request fulfillment."""


class SpecialError(ResponseException):
    """Indicate syntax or spam-prevention issues."""

    def __init__(self, response):
        """Initialize a SpecialError exception instance.

        :param response: A requests.response instance containing a message
        and a list of special errors.

        """
        self.response = response

        resp_dict = self.response.json()  # assumes valid JSON
        self.message = resp_dict.get('message', '')
        self.reason = resp_dict.get('reason', '')
        self.special_errors = resp_dict.get('special_errors', [])
        GlanceDBException.__init__(self, 'Special error {!r}'.format(self.message))


class TooLarge(ResponseException):
    """Indicate that the request data exceeds the allowed limit."""


class UnavailableForLegalReasons(ResponseException):
    """Indicate that the requested URL is unavilable due to legal reasons."""


class UnhandledResponse(ResponseException):
    """Indicate a response status code we have not dealt with yet."""



import attr


def make_supporting_note(data):
    return SupportingNote(id=data['id'], label=data['label'], url=data['url'])


def make_supporting_note_list(data):
    return list(map(make_supporting_note, data))


def make_paper(data):
    return Paper(
        id=data['id'],
        short_title=data['short_title'],
        full_title=data['full_title'],
        pub_short_title=data['pub_short_title'],
        creation_date=data['creation_date'],
        status=data['status'],
        deletion_request=data['deletion_request'],
        deletion_reason=data['deletion_reason'],
        deletion=data['deletion'],
        supporting_notes=make_supporting_note_list(data['supporting_notes'])
        if 'supporting_notes' in data
        else [],
    )


def make_paper_list(data):
    return PaperList(papers=list(map(make_paper, data)))


@attr.s
class SupportingNote(object):
    id = attr.ib()
    label = attr.ib()
    url = attr.ib()


@attr.s
class Paper(object):
    id = attr.ib()
    short_title = attr.ib()
    full_title = attr.ib()
    pub_short_title = attr.ib()
    creation_date = attr.ib()
    status = attr.ib()
    deletion_request = attr.ib()
    deletion_reason = attr.ib()
    deletion = attr.ib()
    supporting_notes = attr.ib()


@attr.s
class PaperList(object):
    papers = attr.ib(type=list)



import attr


def make_phase0(data):
    return Phase0(
        id=data['id'],
        start_date=data['start_date'],
        state=data['state'],
        main_physics_aim=data['main_physics_aim'],
        dataset_used=data['dataset_used'],
        model_tested=data['model_tested'],
        methods=data['methods'],
        editorial_board_formed_date=data['editorial_board_formed_date'],
        group_pre_sign_off_date=data['group_pre_sign_off_date'],
        ana_coord_target_date_comment=data['ana_coord_target_date_comment'],
    )


def make_analysis(data):
    return Analysis(
        id=data['id'],
        short_title=data['short_title'],
        full_title=data['full_title'],
        pub_short_title=data['pub_short_title'],
        creation_date=data['creation_date'],
        status=data['status'],
        deletion_request=data['deletion_request'],
        deletion_reason=data['deletion_reason'],
        deletion=data['deletion'],
        phase_0=make_phase0(data['phase_0']) if 'phase_0' in data else None,
    )


def make_analysis_list(data):
    return AnalysisList(analyses=list(map(make_analysis, data)))


@attr.s
class Phase0(object):
    id = attr.ib()
    start_date = attr.ib()
    state = attr.ib()
    main_physics_aim = attr.ib()
    dataset_used = attr.ib()
    model_tested = attr.ib()
    methods = attr.ib()
    editorial_board_formed_date = attr.ib()
    group_pre_sign_off_date = attr.ib()
    ana_coord_target_date_comment = attr.ib()


@attr.s
class Analysis(object):
    id = attr.ib()
    short_title = attr.ib()
    full_title = attr.ib()
    pub_short_title = attr.ib()
    creation_date = attr.ib()
    status = attr.ib()
    deletion_request = attr.ib()
    deletion_reason = attr.ib()
    deletion = attr.ib()
    phase_0 = attr.ib()


@attr.s
class AnalysisList(object):
    analyses = attr.ib(type=list)



import stare
import betamax


def test_get(auth_session):
    with betamax.Betamax(auth_session).use_cassette('test_paper.test_get'):
        response = auth_session.get('papers')
        assert auth_session._response.status_code == 200
        assert response
        assert 'papers' in response
        assert stare.models.paper.make_paper_list(response['papers'])



import stare
import logging
import betamax

# because expiration will fail (since we cache this information) skip the verification of expiration
jwtOptions = {'verify_signature': False, 'verify_iat': False, 'verify_exp': False}


def test_user_create():
    user = stare.core.User(apiKey='foo')
    assert user.access_token is None
    assert user.bearer == ''
    assert user.id_token == {}
    assert user.name == ''
    assert user.expires_at == 0
    assert user.expires_in == 0
    assert user.is_expired()


def test_user_bad_login(caplog):
    user = stare.core.User(apiKey='foo', jwtOptions=jwtOptions)
    with betamax.Betamax(user._session).use_cassette('test_user.test_user_bad_login'):
        with caplog.at_level(logging.INFO, 'stare'):
            user.authenticate()
            assert 'Invalid key.' in caplog.text
        assert user.is_authenticated() == False
        assert user._response


# NB: because we are using betamax - the ID token which is invalid after 2
# hours is kept so user.is_expired() will be true for testing, do not assert it
def test_user_good_login(caplog):
    user = stare.core.User(jwtOptions=jwtOptions)
    with betamax.Betamax(user._session).use_cassette('test_user.test_user_good_login'):
        with caplog.at_level(logging.INFO, 'stare'):
            user.authenticate()
            assert caplog.text == ''
        assert user.is_authenticated()
        assert user._response



import logging
import sys
import click
import json
import os

from .version import __version__
from . import Glance, settings, utilities

logging.basicConfig(format=utilities.FORMAT_STRING, level=logging.INFO)
log = logging.getLogger(__name__)

client = Glance()


@click.group(context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(version=__version__)
@click.option(
    '--apiKey',
    prompt=not (bool(settings.GLANCE_API_KEY)),
    default=settings.GLANCE_API_KEY,
    show_default=True,
)
@click.option('--site-url', default=settings.SITE_URL, show_default=True)
@click.option(
    '--save-auth',
    help='Filename to save authenticated user to for persistence between requests',
    default='.auth',
)
def stare(apikey, site_url, save_auth):
    global client
    os.environ['GLANCE_API_KEY'] = apikey
    os.environ['SITE_URL'] = site_url
    client.session.user._save_auth = save_auth
    client.session.user._load()


@stare.command()
def authenticate():
    client.session.user.authenticate()
    if client.session.user.is_authenticated():
        click.echo(
            "You have signed in as {}(id={}). Your token expires in {}s.\n\t- permissions: {}\n\t- egroups: {}\n\t- usergroups: {}".format(
                client.session.user.name,
                client.session.user.id,
                client.session.user.expires_in,
                client.session.user.permissions,
                client.session.user.egroups,
                client.session.user.usergroups,
            )
        )


@stare.command()
def list_analyses():
    click.echo(json.dumps(client.analyses, indent=2))
    sys.exit(0)


@stare.command()
def list_papers():
    click.echo(json.dumps(client.papers, indent=2))
    sys.exit(0)



import logging

# The background is set with 40 plus the number of the color, and the foreground with 30
# These are the sequences need to get colored ouput
def _get_color_seq(i):
    COLOUR_SEQ = "\033[1;{0:d}m"
    return COLOUR_SEQ.format(30 + i)


# See: https://stackoverflow.com/questions/287871/print-in-terminal-with-colours
# For additional colours, see: https://stackoverflow.com/questions/15580303/python-output-complex-line-with-floats-coloured-by-value
class colours:
    RESET_SEQ = "\033[0m"
    BOLD_SEQ = "\033[1m"
    UNDERLINE_SEQ = '\033[4m'
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = list(
        map(_get_color_seq, range(8))
    )


BASE_FORMAT_STRING = "[$BOLD%(asctime)s$RESET][%(levelname)-18s]  %(message)s ($BOLD%(filename)s$RESET:%(lineno)d)"
FORMAT_STRING = BASE_FORMAT_STRING.replace("$RESET", "").replace("$BOLD", "")
COLOUR_FORMAT_STRING = BASE_FORMAT_STRING.replace("$RESET", colours.RESET_SEQ).replace(
    "$BOLD", colours.BOLD_SEQ
)


class colouredFormatter(logging.Formatter):
    LEVELS = {
        'WARNING': colours.YELLOW,
        'INFO': colours.WHITE,
        'DEBUG': colours.BLUE,
        'CRITICAL': colours.YELLOW,
        'ERROR': colours.RED,
    }

    def __init__(self, msg, use_colour=True):
        logging.Formatter.__init__(self, msg)
        self.use_colour = use_colour

    def format(self, record):
        levelname = record.levelname
        if self.use_colour and levelname in self.LEVELS:
            levelname_colour = self.LEVELS[levelname] + levelname + colours.RESET_SEQ
            record.levelname = levelname_colour
        return logging.Formatter.format(self, record)


# Custom logger class with multiple destinations
class colouredLogger(logging.Logger):
    def __init__(self, name):
        logging.Logger.__init__(self, name, logging.WARNING)
        colour_formatter = colouredFormatter(COLOUR_FORMAT_STRING)
        console = logging.StreamHandler()
        console.setFormatter(colour_formatter)
        self.addHandler(console)
        return


# Custom logger class with multiple destinations
class BoringLogger(logging.Logger):
    FORMAT = "[%(asctime)s][%(levelname)-18s]  %(message)s (%(filename)s:%(lineno)d)"

    def __init__(self, name):
        logging.Logger.__init__(self, name, logging.WARNING)
        colour_formatter = colouredFormatter(FORMAT_STRING, False)
        console = logging.StreamHandler()
        console.setFormatter(colour_formatter)
        self.addHandler(console)
        return
